use std::sync::Arc;

use glam::{EulerRot, Mat4, Quat, Vec3};
use vulkano::{
    Validated,
    buffer::{AllocateBufferError, Subbuffer},
    memory::allocator::StandardMemoryAllocator,
};

use super::{mvp::MVP, transform::Transform};

#[derive(Default)]
pub struct Camera {
    view: Mat4,
    projection: Mat4,

    position: Vec3,
    rotation: Vec3,
}

impl Camera {
    pub fn new(view: Mat4, projection: Mat4) -> Self {
        Self {
            view,
            projection,
            position: Vec3::ZERO,
            rotation: Vec3::ZERO,
        }
    }

    pub fn rotate(&mut self, rotation: Vec3) {
        self.rotation += rotation;
    }

    pub fn translate(&mut self, translation: Vec3) {
        self.position += translation;
    }

    pub fn set_view(&mut self, view: Mat4) {
        self.view = view;
    }

    pub fn set_projection(&mut self, projection: Mat4) {
        self.projection = projection;
    }

    pub fn get_rotation(&self) -> Vec3 {
        self.rotation
    }

    pub fn set_rotation(&mut self, rotation: Vec3) {
        self.rotation = rotation;
    }

    pub fn get_position(&self) -> Vec3 {
        self.position
    }

    pub fn set_position(&mut self, position: Vec3) {
        self.position = position;
    }

    pub fn create_buffer(
        &self,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[MVP]>, Validated<AllocateBufferError>> {
        let mut world_matrix = Mat4::IDENTITY;
        world_matrix *= Mat4::from_quat(Quat::from_rotation_z(self.rotation.z));
        world_matrix *= Mat4::from_quat(Quat::from_rotation_y(self.rotation.y));
        world_matrix *= Mat4::from_quat(Quat::from_rotation_x(self.rotation.x));
        world_matrix *= Mat4::from_translation(-self.position);

        MVP::new(&world_matrix, &self.view, &self.projection).into_buffer(memory_allocator)
    }
}
